from . import DefaultOpt
from . import resources
from .cloud_meta import get_instance_id
from importlib.metadata import version, PackageNotFoundError
from pulumi.automation import LocalWorkspaceOptions
from pulumi.automation import ProjectBackend
from pulumi.automation import ProjectSettings
from pulumi.automation import create_or_select_stack
from typing import Annotated, Callable, get_type_hints
import click
import copy
import os
import sentry_sdk


def get_installed_package_version(package_name: str) -> str:
    try:
        return version(package_name)
    except PackageNotFoundError:
        return f"Package '{package_name}' is not installed."

instance_id, cloud_provider = get_instance_id()
sentry_sdk.set_context("cloud_metadata", {"instance-id": instance_id, "cloud-provider": cloud_provider})
sentry_sdk.init(release=get_installed_package_version("sparecores-runner"))


def get_stack_name(vendor: str, func: Callable, resource_opts: dict) -> str:
    hints = get_type_hints(func, include_extras=True)
    stack_name = [vendor]
    for name, annotation in hints.items():
        for meta in getattr(annotation, "__metadata__", []):
            # if an option is annotated as StackName, it will be included in the stack name
            if isinstance(meta, resources.StackName):
                stack_name.append(str(resource_opts.get(name)))
    return ".".join(stack_name)


def pulumi_stack(
    pulumi_program: Callable,
    project_name: Annotated[str, DefaultOpt(["--project-name"], type=str, help="Pulumi project name")] = os.environ.get("PULUMI_PROJECT_NAME", "runner"),
    work_dir: Annotated[str, DefaultOpt(["--work-dir"], type=str, help="Pulumi work dir")] = os.environ.get("PULUMI_WORK_DIR", "/data/workdir"),
    pulumi_home: Annotated[str, DefaultOpt(["--pulumi-home"], type=str, help="Pulumi home")] = os.environ.get("PULUMI_HOME", "/root/.pulumi"),
    pulumi_backend_url: Annotated[str, DefaultOpt(["--pulumi-backend-url"], type=str, help="Pulumi backend URL")] = os.environ.get("PULUMI_BACKEND_URL", "file:///data/backend"),
    stack_name: Annotated[
        str,
        click.Option(["--stack-name"], type=str, help="Pulumi stack name, defaults to {vendor}.{region}.{zone}.{instance_id} or similar")]
    = os.environ.get("PULUMI_STACK_NAME", ""),
):
    sentry_sdk.set_context("pulumi", {
        "project_name": project_name,
        "work_dir": work_dir,
        "pulumi_home": pulumi_home,
        "pulumi_backend_url": pulumi_backend_url,
        "stack_name": stack_name,
    })
    stack = create_or_select_stack(
        stack_name=stack_name,
        project_name=project_name,
        program=pulumi_program,
        opts=LocalWorkspaceOptions(
            work_dir=work_dir,
            pulumi_home=pulumi_home,
            project_settings=ProjectSettings(
                name=project_name,
                runtime="python",
                backend=ProjectBackend(pulumi_backend_url)
            )))
    return stack


def create(vendor, pulumi_opts, resource_opts, stack_opts=dict(on_output=print)):
    # don't modify incoming opts
    pulumi_opts = copy.deepcopy(pulumi_opts)
    resource_f = getattr(resources, f"{resources.PREFIX}{vendor}")
    if not pulumi_opts.get("stack_name"):
        pulumi_opts["stack_name"] = get_stack_name(vendor, resource_f, resource_opts)

    def pulumi_program():
        return resource_f(**resource_opts)

    stack = pulumi_stack(pulumi_program, **pulumi_opts)
    stack.up(**stack_opts)


def destroy(vendor, pulumi_opts, resource_opts, stack_opts=dict(on_output=print)):
    # don't modify incoming opts
    pulumi_opts = copy.deepcopy(pulumi_opts)
    resource_f = getattr(resources, f"{resources.PREFIX}{vendor}")
    if not pulumi_opts.get("stack_name"):
        pulumi_opts["stack_name"] = get_stack_name(vendor, resource_f, resource_opts)

    stack = pulumi_stack(lambda: None, **pulumi_opts)
    stack.up(**stack_opts)


def destroy_stack(vendor, pulumi_opts, resource_opts, stack_opts=dict(on_output=print)):
    # don't modify incoming opts
    pulumi_opts = copy.deepcopy(pulumi_opts)
    resource_f = getattr(resources, f"{resources.PREFIX}{vendor}")
    if not pulumi_opts.get("stack_name"):
        pulumi_opts["stack_name"] = get_stack_name(vendor, resource_f, resource_opts)

    stack = pulumi_stack(lambda: None, **pulumi_opts)
    stack.refresh(**stack_opts)
    stack.destroy(**stack_opts)
    stack.workspace.remove_stack(stack.name)


def cancel(vendor, pulumi_opts, resource_opts):
    # don't modify incoming opts
    pulumi_opts = copy.deepcopy(pulumi_opts)
    resource_f = getattr(resources, f"{resources.PREFIX}{vendor}")
    if not pulumi_opts.get("stack_name"):
        pulumi_opts["stack_name"] = get_stack_name(vendor, resource_f, resource_opts)

    stack = pulumi_stack(lambda: None, **pulumi_opts)
    stack.cancel()


def get_stack(vendor, pulumi_opts, resource_opts):
    # don't modify incoming opts
    pulumi_opts = copy.deepcopy(pulumi_opts)
    resource_f = getattr(resources, f"{resources.PREFIX}{vendor}")
    if not pulumi_opts.get("stack_name"):
        pulumi_opts["stack_name"] = get_stack_name(vendor, resource_f, resource_opts)

    return pulumi_stack(lambda: None, **pulumi_opts)



import base64

from .. import DefaultOpt, JSON
from .. import data
from .base import StackName, default, defaults
from typing import Annotated
import click
import copy
import os
import pulumi
import pulumi_gcp as gcp


DEFAULTS = {
    "instance_opts": ("GCP_INSTANCE_OPTS", dict(labels={"created-by": "sc-runner"})),
    "bootdisk_opts": ("GCP_BOOTDISK_OPTS", dict()),
    "bootdisk_init_opts": ("GCP_BOOTDISK_INIT_OPTS", dict(image="ubuntu-2404-lts-amd64")),
    "scheduling_opts": ("GCP_SCHEDULING_OPTS", dict()),
}

def resources_gcp(
        zone: Annotated[str, DefaultOpt(["--zone"], type=click.Choice(data.zones("gcp")), help="Availability zone"), StackName()] = os.environ.get("GCP_ZONE", "us-east1-d"),
        instance: Annotated[str, DefaultOpt(["--instance"], type=click.Choice(data.servers("gcp")), help="Instance type"), StackName()] = os.environ.get("INSTANCE_TYPE", "e2-micro"),
        public_key: Annotated[str, DefaultOpt(["--public-key"], type=str, help="SSH public key")] = os.environ.get("SSH_PUBLIC_KEY", ""),
        instance_opts: Annotated[str, DefaultOpt(["--instance-opts"], type=JSON, default=defaults(DEFAULTS, "instance_opts"), help="Pulumi gcp.compute.Instance options")] = default(DEFAULTS, "instance_opts"),
        bootdisk_opts: Annotated[str, DefaultOpt(["--bootdisk-opts"], type=JSON, default=defaults(DEFAULTS, "bootdisk_opts"), help="Pulumi gcp.compute.InstanceBootDiskArgs options")] = default(DEFAULTS, "bootdisk_opts"),
        bootdisk_init_opts: Annotated[str, DefaultOpt(["--bootdisk-init-opts"], type=JSON, default=defaults(DEFAULTS, "bootdisk_init_opts"), help="Pulumi gcp.compute.InstanceBootDiskInitializeParamsArgs options")] = default(DEFAULTS, "bootdisk_init_opts"),
        scheduling_opts: Annotated[str, DefaultOpt(["--scheduling-opts"], type=JSON, default=defaults(DEFAULTS, "scheduling_opts"), help="Pulumi gcp.compute.InstanceSchedulingArgs options")] = default(DEFAULTS, "scheduling_opts"),
        user_data: Annotated[str | None, DefaultOpt(["--user-data"], type=str, help="Base64 encoded string with user_data script to run at boot")] = os.environ.get("USER_DATA", None),
        disk_size: Annotated[int, DefaultOpt(["--disk-size"], type=int, help="Boot disk size in GiBs")] = int(os.environ.get("DISK_SIZE", 30)),
):
    if "zone" in instance_opts:
        # as zone is part of the Pulumi stack name, it must be specified in the zone option and not in instance_opts
        raise ValueError("zone must be specified in the zone option")
    # we don't want to modify the default
    instance_opts = copy.deepcopy(instance_opts)
    if user_data:
        instance_opts["metadata_startup_script"] = base64.b64decode(user_data)
    if disk_size:
        bootdisk_init_opts["size"] = disk_size

    provider = gcp.Provider(
        resource_name=zone,
        zone=zone,
    )
    if public_key:
        if "metadata" in instance_opts:
            instance_opts["metadata"]["ssh-keys"] = f"ubuntu:{public_key}"
        else:
            instance_opts["metadata"] = {"ssh-keys": f"ubuntu:{public_key}"}
    instance_opts |= dict(
        machine_type=instance,
        zone=zone,
        boot_disk=gcp.compute.InstanceBootDiskArgs(
            initialize_params=gcp.compute.InstanceBootDiskInitializeParamsArgs(**bootdisk_init_opts),
            **bootdisk_opts,
        ),
        network_interfaces=[
            gcp.compute.InstanceNetworkInterfaceArgs(
                network="default",
                access_configs=[gcp.compute.InstanceNetworkInterfaceAccessConfigArgs()]
            )
        ],
    )
    if scheduling_opts:
        instance_opts["scheduling"] = gcp.compute.InstanceSchedulingArgs(**scheduling_opts)
    gcp.compute.Instance(
        instance,
        **instance_opts,
        opts=pulumi.ResourceOptions(provider=provider),
    )


from . import data
from . import resources
from . import runner
from typing import get_type_hints
import click
from click._utils import UNSET
import inspect


def add_click_opts(func):
    def inner(cmd):
        """Convert type hints into click options."""
        hints = get_type_hints(func, include_extras=True)
        for name, annotation in hints.items():
            default = inspect.signature(func).parameters[name].default
            for meta in getattr(annotation, "__metadata__", []):
                if isinstance(meta, (click.Option, click.Argument)):
                    # If the function parameter has a default, use it as the Click option default.
                    # This ensures we don't need to duplicate the default in the DefaultOpt call.
                    #
                    # Click 8.3.0+ behavior change:
                    # In Click 8.2.0 and earlier, when an option didn't have a default set,
                    # the Option.default attribute was None. However, in Click 8.3.0+, the
                    # default attribute is set to Sentinel.UNSET (from click._utils) when not
                    # explicitly provided. This change was made to better distinguish between
                    # "no default set" (UNSET) and "default explicitly set to None".
                    #
                    # The problem this caused:
                    # When we check `if not getattr(meta, "default")`, this evaluates to False
                    # for both None and UNSET (since UNSET is truthy). However, we need to
                    # explicitly check for UNSET to properly detect when a default wasn't set.
                    # Without this check, we wouldn't set the function parameter's default as
                    # the Click option default, causing Click to pass None values through
                    # Context.invoke() (as per Click 8.3.1 changelog: "Replace Sentinel.UNSET
                    # default values by None as they're passed through the Context.invoke()
                    # method"), which would override function defaults when unpacking with **kwargs.
                    #
                    # The fix:
                    # We explicitly check if meta.default is None or UNSET, and if so, we set
                    # it to the function parameter's default value. This ensures that function
                    # defaults are properly used by Click, avoiding the need to duplicate
                    # default values in both the DefaultOpt call and the function parameter.
                    meta_default = getattr(meta, "default", None)
                    if default is not inspect.Parameter.empty:
                        # Only override if meta doesn't have a non-None default explicitly set.
                        # UNSET means it wasn't set, so we should use the function default.
                        # None might also mean it wasn't set (for backward compatibility with
                        # older Click versions or explicit None defaults).
                        if meta_default is None or meta_default is UNSET:
                            meta.default = default
                    if meta.name != name:
                        # in case of a different argument and click option name, replace the
                        # click.Option's name, so it will be mapped correctly
                        # _type: Annotated[str, DefaultOpt(["--type"])
                        meta.name = name
                    cmd.params.append(meta)
        return cmd
    return inner


@click.group()
def cli():
    pass


@add_click_opts(runner.pulumi_stack)
@cli.group()
def create(**kwargs):
    pass


@add_click_opts(runner.pulumi_stack)
@cli.group()
def destroy(**kwargs):
    pass


@add_click_opts(runner.pulumi_stack)
@cli.group()
def destroy_stack(**kwargs):
    """
    Destroy the underlying Pulumi stack.
    Useful if the cloud resources were deleted and the normal destroy command fails.
    """
    pass


@add_click_opts(runner.pulumi_stack)
@cli.group()
def cancel(**kwargs):
    pass


for vendor in data.vendors():
    if vendor not in resources.supported_vendors:
        # exclude not yet supported vendors
        continue

    @create.command(name=vendor)
    @click.pass_context
    def create_resources(ctx, **kwargs):
        pulumi_opts = ctx.parent.params
        vendor = ctx.command.name
        runner.create(vendor, pulumi_opts, kwargs)

    @destroy.command(name=vendor)
    @click.pass_context
    def destroy_resources(ctx, **kwargs):
        pulumi_opts = ctx.parent.params
        vendor = ctx.command.name
        runner.destroy(vendor, pulumi_opts, kwargs)

    @destroy_stack.command(name=vendor)
    @click.pass_context
    def destroy_stack_cmd(ctx, **kwargs):
        pulumi_opts = ctx.parent.params
        vendor = ctx.command.name
        runner.destroy_stack(vendor, pulumi_opts, kwargs)

    @cancel.command(name=vendor)
    @click.pass_context
    def cancel_resources(ctx, **kwargs):
        pulumi_opts = ctx.parent.params
        vendor = ctx.command.name
        runner.cancel(vendor, pulumi_opts, kwargs)

    # add click options from the resource method's annotated argument list
    add_click_opts(getattr(resources, f"{resources.PREFIX}{vendor}"))(create_resources)
    add_click_opts(getattr(resources, f"{resources.PREFIX}{vendor}"))(destroy_resources)
    add_click_opts(getattr(resources, f"{resources.PREFIX}{vendor}"))(destroy_stack_cmd)
    add_click_opts(getattr(resources, f"{resources.PREFIX}{vendor}"))(cancel_resources)


if __name__ == "__main__":
    cli()



from .. import DefaultOpt, JSON
from .. import data
from .base import StackName, default, defaults
from typing import Annotated
import click
import copy
import os
import pulumi
import pulumi_alicloud as alicloud


DEFAULTS = {
    "tags": ("ALICLOUD_TAGS", {"Created-by": "sc-runner"}),
    "instance_opts": ("ALICLOUD_INSTANCE_OPTS", dict(
        internet_charge_type="PayByTraffic",
        internet_max_bandwidth_out=5,
    )),
    "sg_opts": ("ALICLOUD_SG_OPTS", dict()),
    "vpc_opts": ("ALICLOUD_VPC_OPTS", dict(cidr_block="172.16.0.0/12")),
    "vswitch_opts": ("ALICLOUD_VSWITCH_OPTS", dict(cidr_block="172.16.0.0/21")),
}

def resources_alicloud(
        region: Annotated[str, DefaultOpt(["--region"], type=click.Choice(data.regions("alicloud")), help="Region"), StackName()] = os.environ.get("ALIBABA_CLOUD_REGION", "us-west-1"),
        instance: Annotated[str, DefaultOpt(["--instance"], type=click.Choice(data.servers("alicloud")), help="Instance type"), StackName()] = os.environ.get("INSTANCE_TYPE", "ecs.t5-lc1m1.small"),
        public_key: Annotated[str, DefaultOpt(["--public-key"], type=str, help="SSH public key")] = os.environ.get("SSH_PUBLIC_KEY", ""),
        tags: Annotated[str, DefaultOpt(["--tags"], type=JSON, default=defaults(DEFAULTS, "tags"), help="Tags for created resources")] = default(DEFAULTS, "tags"),
        instance_opts: Annotated[str, DefaultOpt(["--instance-opts"], type=JSON, default=defaults(DEFAULTS, "instance_opts"), help="Pulumi alicloud.ecs.Instance options")] = default(DEFAULTS, "instance_opts"),
        sg_opts: Annotated[str, DefaultOpt(["--sg-opts"], type=JSON, default=defaults(DEFAULTS, "sg_opts"), help="Pulumi alicloud.ecs.SecurityGroup options")] = default(DEFAULTS, "sg_opts"),
        vpc_opts: Annotated[str, DefaultOpt(["--vpc-opts"], type=JSON, default=defaults(DEFAULTS, "vpc_opts"), help="Pulumi alicloud.vpc.Network options")] = default(DEFAULTS, "vpc_opts"),
        vswitch_opts: Annotated[str, DefaultOpt(["--vswitch-opts"], type=JSON, default=defaults(DEFAULTS, "vswitch_opts"), help="Pulumi alicloud.vpc.Switch options")] = default(DEFAULTS, "vswitch_opts"),
        image_name: Annotated[str, DefaultOpt(["--image-name"], type=str, help="Image name regex")] = os.environ.get("ALICLOUD_IMAGE_NAME", "^ubuntu_24_04.*20G"),
        user_data: Annotated[str | None, DefaultOpt(["--user-data"], type=str, help="Base64 encoded string with user_data script to run at boot")] = os.environ.get("USER_DATA", None),
        disk_size: Annotated[int, DefaultOpt(["--disk-size"], type=int, help="Boot disk size in GiBs")] = int(os.environ.get("DISK_SIZE", 30)),
):
    # as this function might be called multiple times, and we change the values below, we must make sure we work on copies
    instance_opts = copy.deepcopy(instance_opts)
    sg_opts = copy.deepcopy(sg_opts)
    vpc_opts = copy.deepcopy(vpc_opts)
    vswitch_opts = copy.deepcopy(vswitch_opts)

    provider = alicloud.Provider(
        resource_name=region,
        region=region,
    )

    # Get architecture for image selection
    arch = data.server_cpu_architecture("alicloud", instance).lower().replace("i386", "x86_64")
    if "arm" in arch:
        arch = "arm64"
    else:
        arch = "x86_64"

    # Get Ubuntu image
    if "image_id" not in instance_opts:
        images = alicloud.ecs.get_images(
            owners="system",
            name_regex=image_name,
            architecture=arch,
            opts=pulumi.InvokeOptions(provider=provider),
        )
        if not images.images:
            raise ValueError(f"No image found matching {image_name} for architecture {arch}")
        instance_opts["image_id"] = images.images[0].id

    # Get VPC if not provided
    vpc_id = sg_opts.get("vpc_id") or instance_opts.get("vpc_id")
    vswitch_id = instance_opts.get("vswitch_id")
    created_vpc = None

    if not vpc_id:
        # First try to get the default VPC
        vpc_data = alicloud.vpc.get_networks(
            is_default=True,
            opts=pulumi.InvokeOptions(provider=provider),
        )
        if vpc_data.vpcs:
            vpc_id = vpc_data.vpcs[0].vpc_id
        else:
            # No default VPC, try to get any VPC
            vpc_data = alicloud.vpc.get_networks(
                opts=pulumi.InvokeOptions(provider=provider),
            )
            if vpc_data.vpcs:
                vpc_id = vpc_data.vpcs[0].vpc_id

    # If still no VPC, create one
    if not vpc_id:
        created_vpc = alicloud.vpc.Network(
            instance,
            vpc_name=f"sc-runner-{instance}",
            opts=pulumi.ResourceOptions(provider=provider),
            **vpc_opts,
        )
        vpc_id = created_vpc.id

    # Get a vswitch if not provided
    if not vswitch_id:
        # Only search for existing vswitches if we didn't create the VPC
        if created_vpc is None:
            vswitch_data = alicloud.vpc.get_switches(
                vpc_id=vpc_id,
                opts=pulumi.InvokeOptions(provider=provider),
            )
            if vswitch_data.vswitches:
                vswitch_id = vswitch_data.vswitches[0].vswitch_id

    # If still no vswitch, create one
    if not vswitch_id:
        # Get available zones for VSwitch creation
        zones_data = alicloud.get_zones(
            available_resource_creation="VSwitch",
            opts=pulumi.InvokeOptions(provider=provider),
        )
        if not zones_data.zones:
            raise ValueError(f"No zones available for VSwitch creation in region {region}")

        zone_id = zones_data.zones[0].id
        vswitch_opts["vpc_id"] = vpc_id
        vswitch_opts["zone_id"] = zone_id

        created_vswitch = alicloud.vpc.Switch(
            instance,
            vswitch_name=f"sc-runner-{instance}",
            opts=pulumi.ResourceOptions(provider=provider),
            **vswitch_opts,
        )
        vswitch_id = created_vswitch.id

    # Create key pair if public key is provided
    if public_key and "key_name" not in instance_opts:
        key_pair = alicloud.ecs.EcsKeyPair(
            instance,
            key_pair_name=instance,
            public_key=public_key,
            opts=pulumi.ResourceOptions(provider=provider),
        )
        instance_opts["key_name"] = key_pair.key_pair_name

    # Create security group with allow-all rules
    if vpc_id:
        sg_opts["vpc_id"] = vpc_id
    sg = alicloud.ecs.SecurityGroup(
        instance,
        security_group_name=instance,
        description="Security group created by sc-runner",
        opts=pulumi.ResourceOptions(provider=provider),
        **sg_opts,
    )

    # Add ingress rules (allow all)
    alicloud.ecs.SecurityGroupRule(
        f"{instance}-ingress-v4",
        security_group_id=sg.id,
        type="ingress",
        ip_protocol="all",
        nic_type="intranet",
        policy="accept",
        port_range="-1/-1",
        priority=1,
        cidr_ip="0.0.0.0/0",
        opts=pulumi.ResourceOptions(provider=provider),
    )
    alicloud.ecs.SecurityGroupRule(
        f"{instance}-ingress-v6",
        security_group_id=sg.id,
        type="ingress",
        ip_protocol="all",
        nic_type="intranet",
        policy="accept",
        port_range="-1/-1",
        priority=1,
        ipv6_cidr_ip="::/0",
        opts=pulumi.ResourceOptions(provider=provider),
    )

    # Add egress rules (allow all)
    alicloud.ecs.SecurityGroupRule(
        f"{instance}-egress-v4",
        security_group_id=sg.id,
        type="egress",
        ip_protocol="all",
        nic_type="intranet",
        policy="accept",
        port_range="-1/-1",
        priority=1,
        cidr_ip="0.0.0.0/0",
        opts=pulumi.ResourceOptions(provider=provider),
    )
    alicloud.ecs.SecurityGroupRule(
        f"{instance}-egress-v6",
        security_group_id=sg.id,
        type="egress",
        ip_protocol="all",
        nic_type="intranet",
        policy="accept",
        port_range="-1/-1",
        priority=1,
        ipv6_cidr_ip="::/0",
        opts=pulumi.ResourceOptions(provider=provider),
    )

    # Set user data if provided
    if user_data:
        instance_opts["user_data"] = user_data

    # Set disk size
    if disk_size:
        instance_opts["system_disk_size"] = disk_size

    # Set instance name and tags
    instance_opts["instance_name"] = instance
    instance_opts["tags"] = tags | {"Name": instance}

    # Create the ECS instance
    alicloud.ecs.Instance(
        instance,
        instance_type=instance,
        security_groups=[sg.id],
        vswitch_id=vswitch_id,
        opts=pulumi.ResourceOptions(provider=provider),
        **instance_opts,
    )



import copy
import os
import json


class StackName:
    """Class to signal that an argument's value should be included in the auto-generated Pulumi stack name."""
    pass


def defaults(defaults, opt_name):
    """Return default value from `defaults` for the `opt_name` as a string."""
    # do a copy, so it's not possible to modify the defaults
    envvar, def_val = copy.deepcopy(defaults[opt_name])
    return os.environ.get(envvar, json.dumps(def_val))


def default(defaults, opt_name):
    """Return default value from `defaults` for the `opt_name` as a Python object(dict)."""
    # do a copy, so it's not possible to modify the defaults
    envvar, def_val = copy.deepcopy(defaults[opt_name])
    return json.loads(os.environ.get(envvar, "null")) or def_val




from .. import DefaultOpt, JSON
from .. import data
from .base import StackName, default, defaults
from typing import Annotated
import click
import os
import pulumi_azure_native as azure_native


# defaults for JSON-based options
# key is the option variable name, value is a tuple of env var name and the default value
DEFAULTS = {
    "tags": ("TAGS", {"Created-by": "sc-runner"}),
    "vnet_opts": ("AZURE_VNET_OPTS", dict(address_space=dict(addressPrefixes=["10.0.0.0/16"]))),
    "subnet_opts": ("AZURE_SUBNET_OPTS", dict(address_prefix="10.0.1.0/24")),
    "publicip_opts": ("AZURE_PUBLICIP_OPTS", dict(delete_option="Delete", public_ip_allocation_method="Dynamic")),
}

def resources_azure(
        region: Annotated[str, DefaultOpt(["--region"], type=click.Choice(data.regions("azure")), help="Region"), StackName()] = os.environ.get("AZURE_REGION", "westeurope"),
        zone: Annotated[str, DefaultOpt(["--zone"], type=click.Choice(sorted(set(data.zones("azure")))), help="Availability zone"), StackName()] = os.environ.get("AZURE_ZONE", None),
        image_publisher: Annotated[str, DefaultOpt(["--image-publisher"], type=str, help="VM Image publisher")] = os.environ.get("AZURE_IMAGE_PUBLISHER", "Canonical"),
        # get the list of images in the mcr.microsoft.com/azure-cli Docker image (very slow):
        # az vm image list --publisher Canonical --all
        image_offer: Annotated[str, DefaultOpt(["--image-offer"], type=str, help="VM Image offer")] = os.environ.get("AZURE_IMAGE_OFFER", "ubuntu-24_04-lts"),
        image_sku: Annotated[str, DefaultOpt(["--image-sku"], type=str, help="VM Image SKU (auto-detected for ARM64 instances)")] = os.environ.get("AZURE_IMAGE_SKU", None),
        image_version: Annotated[str, DefaultOpt(["--image-version"], type=str, help="VM Image version")] = os.environ.get("AZURE_IMAGE_VERSION", "latest"),
        instance: Annotated[str, DefaultOpt(["--instance"], type=click.Choice(data.servers("azure")), help="Instance type"), StackName()] = os.environ.get("INSTANCE_TYPE", "Standard_DS1_v2"),
        public_key: Annotated[str, DefaultOpt(["--public-key"], type=str, help="SSH public key")] = os.environ.get("SSH_PUBLIC_KEY", ""),
        tags: Annotated[str, DefaultOpt(["--tags"], type=JSON, default=defaults(DEFAULTS, "tags"), help="Tags for created resources")] = default(DEFAULTS, "tags"),
        vnet_opts: Annotated[str, DefaultOpt(["--vnet-opts"], type=JSON, default=defaults(DEFAULTS, "vnet_opts"), help="Pulumi azure-native.network.VirtualNetwork options")] = default(DEFAULTS, "vnet_opts"),
        subnet_opts: Annotated[str, DefaultOpt(["--subnet-opts"], type=JSON, default=defaults(DEFAULTS, "subnet_opts"), help="Pulumi azure-native.network.Subnet options")] = default(DEFAULTS, "subnet_opts"),
        publicip_opts: Annotated[str, DefaultOpt(["--publicip-opts"], type=JSON, default=defaults(DEFAULTS, "publicip_opts"), help="Pulumi azure-native.network.PublicIPAddress options")] = default(DEFAULTS, "publicip_opts"),
        user_data: Annotated[str | None, DefaultOpt(["--user-data"], type=str, help="Base64 encoded string with user_data script to run at boot")] = os.environ.get("USER_DATA", None),
        disk_size: Annotated[int, DefaultOpt(["--disk-size"], type=int, help="Boot disk size in GiBs")] = int(os.environ.get("DISK_SIZE", 30)),
):
    # Auto-detect image SKU based on instance architecture if not provided
    if image_sku is None:
        arch = data.server_cpu_architecture("azure", instance).lower()
        if "arm" in arch:
            image_sku = "server-arm64"
        else:
            image_sku = "server"

    res_name = f"{region}{zone}{instance}"
    resource_group = azure_native.resources.ResourceGroup(
        instance,
        location=region,
        resource_group_name=res_name,
        tags=tags,
    )

    vnet = azure_native.network.VirtualNetwork(
        instance,
        resource_group_name=resource_group.name,
        location=resource_group.location,
        tags=tags,
        **vnet_opts,
    )

    subnet = azure_native.network.Subnet(
        instance,
        resource_group_name=resource_group.name,
        virtual_network_name=vnet.name,
        **subnet_opts,
    )

    public_ip = azure_native.network.PublicIPAddress(
        instance,
        resource_group_name=resource_group.name,
        location=resource_group.location,
        tags=tags,
        **publicip_opts,
    )

    network_interface = azure_native.network.NetworkInterface(
        instance,
        resource_group_name=resource_group.name,
        location=resource_group.location,
        ip_configurations=[azure_native.network.NetworkInterfaceIPConfigurationArgs(
            name=instance,
            subnet=azure_native.network.SubnetArgs(
                id=subnet.id
            ),
            private_ip_allocation_method=azure_native.network.IPAllocationMethod.DYNAMIC,
            public_ip_address=azure_native.network.PublicIPAddressArgs(
                id=public_ip.id
            ))],
        tags=tags,
    )

    vmopts = dict(
        resource_group_name=resource_group.name,
        location=resource_group.location,
        network_profile=azure_native.compute.NetworkProfileArgs(
            network_interfaces=[azure_native.compute.NetworkInterfaceReferenceArgs(
                id=network_interface.id
            )]
        ),
        hardware_profile=azure_native.compute.HardwareProfileArgs(vm_size=instance),
        os_profile=azure_native.compute.OSProfileArgs(
            computer_name="sc-runner",
            admin_username="ubuntu",
            linux_configuration=azure_native.compute.LinuxConfigurationArgs(
                disable_password_authentication=True,
                ssh=azure_native.compute.SshConfigurationArgs(
                    public_keys=[azure_native.compute.SshPublicKeyArgs(
                        key_data=public_key,
                        path="/home/ubuntu/.ssh/authorized_keys"
                    )
                    ]
                )
            ),
            custom_data=user_data,
        ),
            storage_profile=azure_native.compute.StorageProfileArgs(
                os_disk=azure_native.compute.OSDiskArgs(
                    create_option="FromImage",
                    managed_disk=azure_native.compute.ManagedDiskParametersArgs(
                        storage_account_type="Standard_LRS"
                    ),
                    caching="ReadWrite",
                    disk_size_gb=disk_size,
                ),
                image_reference=azure_native.compute.ImageReferenceArgs(
                    publisher=image_publisher,
                    offer=image_offer,
                    sku=image_sku,
                    version=image_version,
                ),
            ),
            tags=tags,
    )
    if zone is not None:
        vmopts["zones"] = [zone]
    vm = azure_native.compute.VirtualMachine(instance, **vmopts)



import base64

from .. import DefaultOpt, JSON
from .. import data
from .base import StackName, default, defaults
from typing import Annotated
import click
import copy
import os
import pulumi
import pulumi_hcloud as hcloud


DEFAULTS = {
    "instance_opts": ("HCLOUD_INSTANCE_OPTS", dict(labels={"created-by": "sc-runner"})),
}

def resources_hcloud(
        region: Annotated[str, DefaultOpt(["--region"], type=click.Choice(data.regions("hcloud")), help="Region"), StackName()] = os.environ.get("HCLOUD_REGION", "fsn1-dc14"),
        instance: Annotated[str, DefaultOpt(["--instance"], type=click.Choice(data.servers("hcloud")), help="Instance type"), StackName()] = os.environ.get("INSTANCE_TYPE", "cx22"),
        public_key: Annotated[str, DefaultOpt(["--public-key"], type=str, help="SSH public key")] = os.environ.get("SSH_PUBLIC_KEY", ""),
        instance_opts: Annotated[str, DefaultOpt(["--instance-opts"], type=JSON, default=defaults(DEFAULTS, "instance_opts"), help="Pulumi hcloud.Server options")] = default(DEFAULTS, "instance_opts"),
        user_data: Annotated[str | None, DefaultOpt(["--user-data"], type=str, help="Base64 encoded string with user_data script to run at boot")] = os.environ.get("USER_DATA", None),
):
    # we don't want to modify the default
    instance_opts = copy.deepcopy(instance_opts)

    if public_key and "ssh_keys" not in instance_opts:
        ssh_key = hcloud.SshKey(
            instance,
            name=instance,
            public_key=public_key
        )
        instance_opts["ssh_keys"] = [ssh_key.id]

    hcloud.Server(
        instance,
        name=instance,
        image="ubuntu-24.04",
        server_type=instance,
        datacenter=region,
        user_data=user_data,
        **instance_opts,
    )


from .. import DefaultOpt, JSON
from .. import data
from .base import StackName, default, defaults
from typing import Annotated
import click
import copy
import os
import pulumi
import pulumi_aws as aws


V4_ALLOW_ALL = dict(ip_protocol="-1",
                    cidr_ipv4="0.0.0.0/0",
                    from_port=0,
                    to_port=0,
                    )
V6_ALLOW_ALL = dict(ip_protocol="-1",
                    cidr_ipv6="::/0",
                    from_port=0,
                    to_port=0,
                    )

# defaults for JSON-based options
# key is the option variable name, value is a tuple of env var name and the default value
DEFAULTS = {
    "tags": ("TAGS", {"Created-by": "sc-runner"}),
    "instance_opts": ("AWS_INSTANCE_OPTS", dict(associate_public_ip_address=True)),
    "vpc_opts": ("AWS_VPC_OPTS", dict()),
    "subnet_opts": ("AWS_SUBNET_OPTS", dict()),
    "sg_opts": ("AWS_SG_OPTS", dict()),
    "ingress_rules": ("AWS_INGRESS_RULES", [V4_ALLOW_ALL, V6_ALLOW_ALL]),
    "egress_rules": ("AWS_EGRESS_RULES", [V4_ALLOW_ALL, V6_ALLOW_ALL]),
}


def resources_aws(
        region: Annotated[str, DefaultOpt(["--region"], type=click.Choice(data.regions("aws")), help="Region"), StackName()] = os.environ.get("AWS_DEFAULT_REGION", "us-east-1"),
        zone: Annotated[str, DefaultOpt(["--zone"], type=click.Choice(data.zones("aws")), help="Availability zone"), StackName()] = os.environ.get("AWS_ZONE", None),
        assume_role_arn: Annotated[str, DefaultOpt(["--assume-role-arn"], type=str, help="Role to be assumed")] = os.environ.get("AWS_ASSUME_ROLE_ARN", ""),
        ami_owner: Annotated[str, DefaultOpt(["--ami-owner"], type=str, help="AMI owner")] = os.environ.get("AWS_AMI_OWNER", "099720109477"),
        # to get the available image names:
        # aws ec2 describe-images --region us-east-1 --owners 099720109477 | jq '.Images[].Name'
        ami_name: Annotated[str, DefaultOpt(["--ami-name"], type=str, help="AWS name filter")] = os.environ.get("AWS_AMI_NAME", "ubuntu/images/hvm-ssd-gp3/ubuntu-noble-24.04-*-server*"),
        instance: Annotated[str, DefaultOpt(["--instance"], type=click.Choice(data.servers("aws")), help="Instance type"), StackName()] = os.environ.get("INSTANCE_TYPE", "t3.micro"),
        public_key: Annotated[str, DefaultOpt(["--public-key"], type=str, help="SSH public key")] = os.environ.get("SSH_PUBLIC_KEY", ""),
        tags: Annotated[str, DefaultOpt(["--tags"], type=JSON, default=defaults(DEFAULTS, "tags"), help="Tags for created resources")] = default(DEFAULTS, "tags"),
        instance_opts: Annotated[str, DefaultOpt(["--instance-opts"], type=JSON, default=defaults(DEFAULTS, "instance_opts"), help="Pulumi aws.ec2.Instance options")] = default(DEFAULTS, "instance_opts"),
        vpc_opts: Annotated[str, DefaultOpt(["--vpc-opts"], type=JSON, default=defaults(DEFAULTS, "vpc_opts"), help="Pulumi aws.ec2.Vpc options")] = default(DEFAULTS, "vpc_opts"),
        subnet_opts: Annotated[str, DefaultOpt(["--subnet-opts"], type=JSON, default=defaults(DEFAULTS, "subnet_opts"), help="Pulumi aws.ec2.Subnet options")] = default(DEFAULTS, "subnet_opts"),
        sg_opts: Annotated[str, DefaultOpt(["--sg-opts"], type=JSON, default=defaults(DEFAULTS, "sg_opts"), help="Pulumi aws.ec2.SecurityGroup options")] = default(DEFAULTS, "sg_opts"),
        ingress_rules: Annotated[str, DefaultOpt(["--ingress-rules"], type=JSON, default=defaults(DEFAULTS, "ingress_rules"), help="List of Pulumi aws.ec2.SecurityGroupIngressRule options")] = default(DEFAULTS, "ingress_rules"),
        egress_rules: Annotated[str, DefaultOpt(["--egress-rules"], type=JSON, default=defaults(DEFAULTS, "egress_rules"), help="List of Pulumi aws.ec2.SecurityGroupEgressRule options")] = default(DEFAULTS, "egress_rules"),
        user_data: Annotated[str | None, DefaultOpt(["--user-data"], type=str, help="Base64 encoded string with user_data script to run at boot")] = os.environ.get("USER_DATA", None),
        disk_size: Annotated[int, DefaultOpt(["--disk-size"], type=int, help="Boot disk size in GiBs")] = int(os.environ.get("DISK_SIZE", 30)),
):
    # as this function might be called multiple times, and we change the values below, we must make sure we work on copies
    instance_opts = copy.deepcopy(instance_opts)
    vpc_opts = copy.deepcopy(vpc_opts)
    subnet_opts = copy.deepcopy(subnet_opts)
    sg_opts = copy.deepcopy(sg_opts)
    prov_kwargs = {}
    if assume_role_arn:
        prov_kwargs["assume_role"] = aws.ProviderAssumeRoleArgs(role_arn=assume_role_arn)
    provider = aws.Provider(
        resource_name=region,
        region=region,
        skip_metadata_api_check=False,  # enable instance roles
        default_tags=aws.ProviderDefaultTagsArgs(tags=tags | {"Name": instance}),
        **prov_kwargs,
    )

    if public_key:
        pubkey = aws.ec2.KeyPair(
            instance,
            public_key=public_key,
            key_name=instance,
        )
        instance_opts["key_name"] = pubkey.id
    if user_data:
        instance_opts["user_data_base64"] = user_data
    if disk_size:
        instance_opts["root_block_device"] = aws.ec2.InstanceRootBlockDeviceArgs(volume_size=disk_size)

    if "ami" not in instance_opts:
        # some instances are marked as i386, but they aren't IA-32, replace them, so we can find AMIs
        arch = data.server_cpu_architecture("aws", instance).lower().replace("i386", "x86_64")
        ami = aws.ec2.get_ami(
            most_recent=True,  # in case of a filter is given as the name
            filters=[
                aws.ec2.GetAmiFilterArgs(name="architecture", values=[arch]),
                aws.ec2.GetAmiFilterArgs(name="name", values=[ami_name]),
                aws.ec2.GetAmiFilterArgs(name="virtualization-type", values=["hvm"]),
            ],
            owners=[ami_owner],
            opts=pulumi.InvokeOptions(provider=provider),
        )
        instance_opts["ami"] = ami.id

    # If any of these are given, we assume that the required IDs are set everywhere and there's a working VPC/subnet/routing
    vpc_id = subnet_opts.get("vpc_id") or sg_opts.get("vpc_id")
    subnet_id = instance_opts.get("subnet_id")

    # If vpc_opts is given, we create a VPC/subnet/routing nevertheless
    if vpc_opts:
        vpc = aws.ec2.Vpc(instance, **vpc_opts)
        vpc_id = vpc.id

        subnet_opts["vpc_id"] = vpc.id
        subnet = aws.ec2.Subnet(instance, **subnet_opts)
        subnet_id = subnet.id

        igw = aws.ec2.InternetGateway(
            instance,
            vpc_id=vpc.id,
            opts=pulumi.ResourceOptions(provider=provider),
        )

        rt = aws.ec2.RouteTable(
            instance,
            vpc_id=vpc.id,
            routes=[
                aws.ec2.RouteTableRouteArgs(
                    cidr_block="0.0.0.0/0",
                    gateway_id=igw.id,
                ),
                aws.ec2.RouteTableRouteArgs(
                    ipv6_cidr_block="::/0",
                    gateway_id=igw.id,
                ),
            ],
            opts=pulumi.ResourceOptions(provider=provider),
        )
        aws.ec2.RouteTableAssociation(
            instance,
            subnet_id=subnet.id,
            route_table_id=rt.id,
            opts=pulumi.ResourceOptions(provider=provider),
        )

    if sg_opts or ingress_rules or egress_rules:
        sg_opts["vpc_id"] = vpc_id
        sg = aws.ec2.SecurityGroup(
            instance,
            opts=pulumi.ResourceOptions(provider=provider),
            **sg_opts,
        )
        instance_opts["vpc_security_group_ids"] = [sg.id]
        for i in range(len(ingress_rules)):
            aws.vpc.SecurityGroupIngressRule(
                f"{instance}-{i}",
                security_group_id=sg.id,
                opts=pulumi.ResourceOptions(provider=provider),
                **ingress_rules[i]
            )
        for i in range(len(egress_rules)):
            aws.vpc.SecurityGroupEgressRule(
                f"{instance}-{i}",
                security_group_id=sg.id,
                opts=pulumi.ResourceOptions(provider=provider),
                **egress_rules[i]
            )

    instance_opts["subnet_id"] = subnet_id
    aws.ec2.Instance(
        instance,
        instance_type=instance,
        opts=pulumi.ResourceOptions(provider=provider),
        **instance_opts,
    )



import requests

# Define metadata service endpoints for each cloud provider
CLOUD_METADATA_ENDPOINTS = {
    "aws": "http://169.254.169.254/latest/meta-data/instance-id",
    "azure": "http://169.254.169.254/metadata/instance/compute/vmId?api-version=2021-02-01",
    "gcp": "http://metadata.google.internal/computeMetadata/v1/instance/id",
    "alicloud": "http://100.100.100.200/latest/meta-data/instance-id",
    "digitalocean": "http://169.254.169.254/metadata/v1/id",
    "oracle": "http://169.254.169.254/opc/v1/instance/id",
    "openstack": "http://169.254.169.254/openstack/latest/meta_data.json",
}

# Headers required for specific cloud providers
CLOUD_METADATA_HEADERS = {
    "gcp": {"Metadata-Flavor": "Google"},
    "azure": {"Metadata": "true"},
}


def check_endpoint(url, headers=None):
    try:
        response = requests.get(url, headers=headers, timeout=1)
        if response.status_code == 200:
            return response
    except requests.exceptions.RequestException:
        pass
    return None

def get_instance_id():
    for cloud, url in CLOUD_METADATA_ENDPOINTS.items():
        headers = CLOUD_METADATA_HEADERS.get(cloud, {})
        response = check_endpoint(url, headers)
        if response:
            if cloud == "openstack":
                # OpenStack returns JSON, extract instance ID accordingly
                data = response.json()
                return data.get("uuid"), "openstack"
            else:
                return response.text.strip(), cloud
    return None, None

if __name__ == "__main__":
    instance_id, cloud_provider = get_instance_id()
    if instance_id:
        print(f"Cloud Provider: {cloud_provider.upper()}")
        print(f"Instance ID: {instance_id}")
    else:
        print("Could not determine the cloud environment or fetch the instance ID.")
